/* definicia prenasanych hodnot */
%union  {
    char* meno;
    char* txt;
    char* sig;
    unsigned int mod;
    int ihod;
    double dhod;
}

/* definicia term. symbolov a hodnoty nimi prenasane */
%token          BOOLEAN  DOUBLE  INT  LONG  VOID
%token          CLASS  EXTENDS  CONSTRUCTOR IMPLEMENTS IMPORT
%token          PUBLIC PRIVATE PROTECTED
%token          STATIC FINAL ABSTRACT NATIVE
%token          NEW THIS SUPER RETURN IF ELSE WHILE
%token <meno>   ID
%token <txt>    STR_LITERAL
%token <ihod>   TRIEDA INT_LITERAL
%token <dhod>   DBL_LITERAL
%token          NULL_LITERAL
%token          TRUE_LITERAL
%token          FALSE_LITERAL
%token          LZZAT  PZZAT  LOZAT  POZAT  LHZAT  PHZAT
%token          CIARKA  BCIARKA  BODKA  ROVNE
/* priradenie priorit a asociativnosti operatorom */
%left       AND OR
%nonassoc   LT LE GT GE EQ NE
%left       PLUS MINUS
%left       KRAT DEL

/* priradenie prenasanych hodnot neterm. symbolom */
%type <ihod> NadTrieda
%type <sig>  NepParametre Parametre TypVoid TypDat
%type <mod>  ModTriedy ModClena Mod1 Mod2


%{  /* Udajove struktury pouzivane pri preklade */
typedef unsigned int uint;

/* Datovy clen */
typedef struct {
    uint pristup; /* Modifikatory pristupu */
    char *meno;
    char *signatura;
} dclen;

/* Metoda */
typedef struct {
    uint pristup;     /* Modifikatory pristupu */
    char *meno;
    char *signaturaP; /* Signatura parametrov */
    char *signaturaR; /* Signatura navratovej hodnoty */
    int velkostLP;    /* Velkost pamate pre lokalne premenne */
    char *kod;        /* Kod metody */
} met;

struct {
    char *menoTr;     /* Meno triedy uvedene v kode */
    char *javaTr;     /* Uplne meno triedy pre JVM */
    char *sigTr;      /* Signatura triedy */
    char *menoNadTr;  /* Meno nadtriedy */
    int ixntr;        /* Index na nadtriedu v tabulke tried */
    uint pristup;     /* Modifikatory pristupu */
    int pocDatClenov; /* Pocet datovych clenov */
    dclen* DatCleny;  /* Pole informacii o datovych clenoch */
    int pocMetod;     /* Pocet metod */
    met* Metody;      /* Pole informacii o metodach */
} triedy[20];

int ix_tr = -1; /* Index aktualne spracovavanej triedy */
int ix_m  = -1; /* Index aktualne spracovavanej metody */

/* Modifikatory su ulozene pomcou bitovej masky */
#define MOD_PUBLIC    0x0001
#define MOD_PRIVATE   0x0002
#define MOD_PROTECTED 0x0004
#define MOD_STATIC    0x0008
#define MOD_FINAL     0x0010
#define MOD_NATIVE    0x0100
#define MOD_ABSTRACT  0x0400

    /* prototypy */
int yyparse(void), yylex(void);
void ulozImp(char *,char *);
void ulozTr(char *,int,uint);
void ulozDC(char *,char *,uint);
void ulozMet(char *,char *,char *,uint);
void yyerror(char*), xxerror(char*, char*);
void vypisTriedy(int, char *);

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

char sigpom[80];

/* vlozenie lexikalneho analyzatora */
#include "lex.yy.c"
%}

%%
Subor   : Importy Trieda
        | Subor  Trieda
        ;
Importy : /* e */
        | Importy  Import
        ;
Import  : IMPORT  ID  NATIVE  STR_LITERAL
            { ulozImp($2, $4); }
        ;
Trieda  : CLASS  ModTriedy ID  NadTrieda  ImpRozhrania
            { ulozTr($3, $4, $2); }
          LZZAT  Cleny  PZZAT
        ;
ModTriedy :        { $$ = 0; }
        | PUBLIC   { $$ = MOD_PUBLIC; }
        | ABSTRACT { $$ = MOD_ABSTRACT; }
        | FINAL    { $$ = MOD_FINAL; }
        ;
NadTrieda : /* e */
            { $$ = 0; }
        | EXTENDS  TRIEDA
            { $$ = $2; }
        ;
ImpRozhrania : /* e */
        | ImpRozhrania IMPLEMENTS TRIEDA
        ;
Cleny   : /* e */
        | Cleny DatovyClen
        | Cleny Metoda
        | Cleny error
        ;
DatovyClen : ModClena TypDat  ID  BCIARKA
            { ulozDC($3, $2, $1); }
        ;
ModClena : Mod1 Mod2     { $$ = $1 | $2; }
        ;
Mod1    :           { $$ = 0; }
        | PUBLIC    { $$ = MOD_PUBLIC; }
        | PROTECTED { $$ = MOD_PROTECTED; }
        | PRIVATE   { $$ = MOD_PRIVATE; }
        ;
Mod2    :           { $$ = 0; }
        | STATIC    { $$ = MOD_STATIC; }
        | ABSTRACT  { $$ = MOD_ABSTRACT; }
        | FINAL     { $$ = MOD_FINAL; }
        | STATIC FINAL  { $$ = MOD_STATIC | MOD_FINAL; }
        ;
Metoda  : ModClena TypDat  ID  {;} LOZAT  NepParametre  POZAT
            { ulozMet($3, $6, $2, $1); }
            TeloMet
        | ModClena TypVoid  ID  {;} LOZAT  NepParametre  POZAT
            { ulozMet($3, $6, $2, $1); }
            TeloMet
        | CONSTRUCTOR {;} LOZAT NepParametre POZAT
            { ulozMet("<init>", $4, "V", MOD_PUBLIC); }
            TeloMet
        ;
NepParametre : /* e */
            { $$ = (char*) malloc(3); strcpy($$, "()"); }
        | Parametre
            { $$ = strcat($1, ")"); }
        ;
Parametre : TypDat  ID
            { $$ = (char*) malloc(256); strcpy($$, "("); strcat($$, $1); }
        | Parametre  CIARKA  TypDat  ID
            { $$ = strcat($1, $3); }
        ;
TypVoid : VOID
            { $$ = "V"; }
        ;
TypDat  : INT
            { $$ = "I"; }
        | DOUBLE
            { $$ = "D"; }
        | BOOLEAN
            { $$ = "Z"; }
        | TRIEDA
            { $$ = triedy[$1].sigTr; }
        | TRIEDA LHZAT PHZAT
            { $$ = sigpom; sprintf(sigpom, "[%s", triedy[$1].sigTr); }
        ;
TeloMet : BCIARKA
        | LZZAT  Prikazy  PZZAT
        ;
Prikazy : /* e */
        | Prikazy Prikaz
        | Prikazy error
        ;
Prikaz  : TypDat ZozIdent BCIARKA
        | VyrazPr BCIARKA
        | RETURN BCIARKA
        | BCIARKA
        | RETURN Vyraz BCIARKA
        | WHILE LOZAT Vyraz POZAT Prikaz
        | IF LOZAT Vyraz POZAT Prikaz ELSE Prikaz
        | IF LOZAT Vyraz POZAT Prikaz
        | LZZAT Prikazy PZZAT
        ;
ZozIdent : ID
        | ZozIdent CIARKA ID
        ;
Vyraz   : Primarny
        | Vyraz PLUS Vyraz
        | Vyraz MINUS Vyraz
        | Vyraz KRAT Vyraz
        | Vyraz DEL Vyraz
        | Vyraz LT Vyraz
        | Vyraz LE Vyraz
        | Vyraz GT Vyraz
        | Vyraz GE Vyraz
        | Vyraz EQ Vyraz
        | Vyraz NE Vyraz
        | Vyraz OR Vyraz
        | Vyraz AND Vyraz
        ;
VyrazPr : ID ROVNE Vyraz
        | DCPristup ROVNE Vyraz
        | MetVolanie
        | LOZAT VOID POZAT MetVolanie
        ;
Primarny : ID
        | THIS
        | NovyObjekt
        | MetVolanie
        | DCPristup
        | NULL_LITERAL
        | INT_LITERAL
        | DBL_LITERAL
        | STR_LITERAL
        | TRUE_LITERAL
        | FALSE_LITERAL
        | LOZAT  Vyraz  POZAT
        | TRIEDA LOZAT Vyraz POZAT
        | INT LOZAT Vyraz POZAT
        | DOUBLE LOZAT Vyraz POZAT
        | LONG LOZAT Vyraz POZAT
        ;
NovyObjekt : NEW  TRIEDA  LOZAT  NepArgumenty  POZAT
        ;
MetVolanie : TRIEDA BODKA ID LOZAT NepArgumenty POZAT
        | SUPER LOZAT NepArgumenty POZAT
        | Primarny BODKA ID  LOZAT  NepArgumenty POZAT
        ;
DCPristup : TRIEDA BODKA ID
        | Primarny BODKA ID
        ;
NepArgumenty : /* e */
        | Argumenty
        ;
Argumenty : Vyraz
        | Argumenty  CIARKA  Vyraz
        ;
%%

static int pocetChyb = 0;

void yyerror(char *s)
{
    fprintf(stdout, "CHYBA: %s [%s:%d]\n", s, yytext, riadok);
    pocetChyb++;
}

void xxerror(char *s1, char *s2)
{
    fprintf(stdout, "CHYBA: %s %s [%d]\n", s1, s2, riadok);
    pocetChyb++;
}


int main(int argc, char *argv[])
{
    extern FILE *yyin;
    int i;
    char src[20]; /* Nazov vstupneho suboru */
    if (argc > 1) {
        yyin = fopen(argv[1], "r");
        strcpy(src, argv[1]);
        if (yyin == NULL) {
            yyin = stdin;
            strcpy(src, "stdin");
        }
    }
    ulozImp("Object", "java/lang/Object");
    ulozImp("String", "java/lang/String");
    ulozImp("System", "java/lang/System");
    (void) yyparse();
    fprintf(stderr, "POCET CHYB: %d\n", pocetChyb);
    for (i = 0; i <= ix_tr; i++)
        vypisTriedy(i, src);
    getchar();
    return 0;
}


/*
 * Nacitanie informacii o triede zo standardnej kniznice Java
 * meno - meno triedy
 * jmeno - uplne meno triedy
 * Informacie sa citaju zo suboru "java.api"
 */
void ulozImp(char *meno, char *jmeno)
{
    FILE *fi;
    char tok[80], pmeno[80]; /* token a precitane meno */
    uint pristup;
    fi = fopen("java.api", "r");
    if (fi == NULL) {
        xxerror("Nemozno najst archiv", "java.api");
        return;
    }
    fscanf(fi, "%s", tok);
    while (strcmp(tok, ".end") != 0) {
        if (strcmp(tok, ".class") != 0 && strcmp(tok, ".interface") != 0) {
            fscanf(fi, "%s", tok);
            continue;
        }
        fscanf(fi, "%x %s", &pristup, pmeno);
        if (strcmp(pmeno, jmeno) != 0) {
            fscanf(fi, "%s", tok);
            continue;
        }
        /* Zaciatok zaznamenavania triedy */
        ix_tr++;
        triedy[ix_tr].pristup = pristup | MOD_NATIVE;
        triedy[ix_tr].menoTr = meno;
        triedy[ix_tr].javaTr = jmeno;
        triedy[ix_tr].sigTr = (char*) malloc(strlen(jmeno)+3);
        sprintf(triedy[ix_tr].sigTr, "L%s;", jmeno);
        triedy[ix_tr].menoNadTr = NULL;
        triedy[ix_tr].ixntr = -1;
        fscanf(fi, "%s", tok);
        /* Nadtrieda */
        if (strcmp(tok, ".super") == 0) {
            fscanf(fi, "%s", pmeno);
            for (int i = 0; i < ix_tr; i++) {
                if (strcmp(triedy[i].menoTr, pmeno) == 0 ||
                        strcmp(triedy[i].javaTr, pmeno) == 0) {
                    triedy[ix_tr].menoNadTr = triedy[i].menoTr;
                    triedy[ix_tr].ixntr = i;
                }
            }
            if (triedy[ix_tr].ixntr < 0) {
                xxerror("Nemozno najst nadtriedu:", pmeno);
            }
        }
        /* Datove cleny */
        int pdc; /* pocet datovych clenov */
        fscanf(fi, "%d\n", &pdc);
        triedy[ix_tr].pocDatClenov = pdc;
        if (pdc > 0)
            triedy[ix_tr].DatCleny = (dclen*) calloc(pdc, sizeof(dclen));
        for (int i = 0; i < pdc; i++) {
            char sig[80];
            fscanf(fi, ".field %x %s %s\n", &pristup, pmeno, sig);
            triedy[ix_tr].DatCleny[i].meno = strdup(pmeno);
            triedy[ix_tr].DatCleny[i].signatura = strdup(sig);
            triedy[ix_tr].DatCleny[i].pristup = pristup;
        }
        /* Metody */
        int pm; /* pocet metod */
        fscanf(fi, "%d\n", &pm);
        triedy[ix_tr].pocMetod = pm;
        if (pm > 0)
            triedy[ix_tr].Metody = (met*) calloc(pm, sizeof(met));
        for (int i = 0; i < pm; i++) {
            char sigP[80], sigR[80];
            fscanf(fi, ".method %x %s %s %s\n", &pristup, pmeno, sigP, sigR);
            triedy[ix_tr].Metody[i].pristup=pristup;
            triedy[ix_tr].Metody[i].meno=strdup(pmeno);
            triedy[ix_tr].Metody[i].signaturaP=strdup(sigP);
            triedy[ix_tr].Metody[i].signaturaR=strdup(sigR);
        }
        fclose(fi);
        return;
    }
    fclose(fi);
    xxerror("Nemozno najst JAVA triedu:", jmeno);
}


/*
 * Ulozenie informacii o triede do tabulky tried
 * meno - meno triedy
 * ixNT - index nadtriedy (v tabulke tried)
 * mod  - modifikatory
 */
void ulozTr(char *meno, int ixNT, uint mod)
{
    int i, j, k;
    ix_tr++;
    triedy[ix_tr].menoTr = meno;
    triedy[ix_tr].javaTr = meno;
    triedy[ix_tr].sigTr = (char*) malloc(strlen(meno) + 3);
    sprintf(triedy[ix_tr].sigTr, "L%s;", meno);
    triedy[ix_tr].ixntr = ixNT;
    if (triedy[ixNT].pristup & MOD_FINAL)
        xxerror("Nemozno dedit z FINAL triedy", triedy[ixNT].menoTr);
    triedy[ix_tr].menoNadTr = triedy[ixNT].menoTr;
    triedy[ix_tr].pristup = mod;
    /* Alokacia pamate pre informacie o dat. clenoch a metodach */
    triedy[ix_tr].pocDatClenov = 0;
    triedy[ix_tr].DatCleny = (dclen*) calloc(20, sizeof(dclen));
    triedy[ix_tr].pocMetod = 0;
    triedy[ix_tr].Metody = (met*) calloc(20, sizeof(met));
}


/*
 * Ulozenie informacii o datovom clene triedy
 * meno - meno datoveho clena
 * sig  - signatura (typ)
 * mod  - modifikatory
 */
void ulozDC(char *meno, char *sig, uint mod)
{
    int i, poc;
    poc = triedy[ix_tr].pocDatClenov;
    /* Test, ci datovy clen s danym menom nie je uz deklarovany */
    for (i = 0; i < poc; i++) {
        if (strcmp(triedy[ix_tr].DatCleny[i].meno, meno) == 0) {
            xxerror("Viacnasobna deklaracia datoveho clena:", meno);
            return;
        }
    }
    triedy[ix_tr].DatCleny[poc].meno = meno;
    triedy[ix_tr].DatCleny[poc].pristup = mod;
    if (mod & MOD_ABSTRACT)
        xxerror("Datovy clen nemoze byt abstraktny:", meno);
    triedy[ix_tr].DatCleny[poc].signatura = sig;
    triedy[ix_tr].pocDatClenov++;
}


/*
 * Ulozenie informacii o metode
 * meno - meno metody
 * sigP - signatura parametrov
 * sigR - signatura navratovej hodnoty
 * mod  - modifikatory
 */
void ulozMet(char *meno, char *sigP, char *sigR, uint mod)
{
    int i;
    char *sig;
    ix_m = triedy[ix_tr].pocMetod;
    /* Kontrola jedinecnosti mena metody */
    for (i = 0; i < ix_m; i++) {
        if (strcmp(triedy[ix_tr].Metody[i].meno, meno) == 0 &&
                strcmp(triedy[ix_tr].Metody[i].signaturaP, sigP) == 0) {
            xxerror("Viacnasobna deklaracia metody:", meno);
            return;
        }
    }
    triedy[ix_tr].Metody[ix_m].meno = meno;
    triedy[ix_tr].Metody[ix_m].pristup = mod;
    triedy[ix_tr].Metody[ix_m].signaturaP = sigP;
    triedy[ix_tr].Metody[ix_m].signaturaR = sigR;
    triedy[ix_tr].Metody[ix_m].kod = NULL;
    triedy[ix_tr].pocMetod++;
}


/*
 * Textova podoba modifikatorov
 * mod - modifikatory v podobe bitovej masky
 */
char *vratMod(uint mod)
{
    static char textmod[80];
    textmod[0] = '\0';
    if (mod & MOD_PUBLIC)    strcat(textmod, "public ");
    if (mod & MOD_PROTECTED) strcat(textmod, "protected ");
    if (mod & MOD_PRIVATE)   strcat(textmod, "private ");
    if (mod & MOD_STATIC)    strcat(textmod, "static ");
    if (mod & MOD_FINAL)     strcat(textmod, "final ");
    if (mod & MOD_ABSTRACT)  strcat(textmod, "abstract ");
    if (mod & MOD_NATIVE)    strcat(textmod, "native ");
    return textmod;
}


/*
 * Vypis prelozenej podoby triedy do Jal soboru
 * ix - index triedy v tabulke tried
 * src - nazov suboru so zdrojovym textom
 * Nazov vystupneho suboru je odvodeny na zaklade mena triedy.
 */
void vypisTriedy(int ix, char *src)
{
    int i, ixn;
    char mt[80]; /* Nazov suboru */
    FILE *vyst;
    if (triedy[ix].pristup & MOD_NATIVE)
        return;
    strcpy(mt, triedy[ix].menoTr);
    strcat(mt, ".jal");
    vyst = fopen(mt, "w");
    if (vyst == NULL)
        xxerror("Nemozno vytvorit subor:", mt);

    /* Vypis informacii o triede */
    fprintf(vyst, ".source %s\n", src);
    fprintf(vyst, ".class %s %s\n", vratMod(triedy[ix].pristup), triedy[ix].menoTr);
    ixn = triedy[ix].ixntr; /* nadtrieda */
    fprintf(vyst, ".super %s\n", triedy[ixn].javaTr);
    /* Vypis datovych clenov */
    for (i = 0; i < triedy[ix].pocDatClenov; i++) {
        fprintf(vyst, ".field %s %s %s\n",
            vratMod(triedy[ix].DatCleny[i].pristup),
            triedy[ix].DatCleny[i].meno,
            triedy[ix].DatCleny[i].signatura);
    }
    /* Vypis metod */
    for (i = 0; i < triedy[ix].pocMetod; i++) {
        fprintf(vyst, ".method %s %s%s%s\n",
            vratMod(triedy[ix].Metody[i].pristup),
            triedy[ix].Metody[i].meno,
            triedy[ix].Metody[i].signaturaP,
            triedy[ix].Metody[i].signaturaR);
    }
    fclose(vyst);
}

